//Audio sampling program by R.Welbourn 1998
//Final year Bsc (Hons) Computing and Informatics project

unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Gauges, StdCtrls, ComCtrls, MMSystem, MPlayer, Menus, ExtCtrls, Fourier, Math,
  DB, DBTables, Meter, RChart;

const
  SampleRate = 11025; //Sample rate of soundcard
  BufferLength = 2048;  //Length of audio buffer - buffer size / sample rate =
                       //buffer length in seconds.
  BitsPerSample = 8;
  SampleRange = 255;
  SpectrumSize=BufferLength;
  Noise=200;
  Sweet=500;
  Threshold=15;
  CutOff=210;

type
  AudioSample = 0..SampleRange; //In 8 bit sampling the values are integers from
                         //0 to 255 where 127 is silence, 255 is max positive
                         //amplitude and 0 is max -ve amplitude.

type
  tBuffer = array [0..BufferLength] of AudioSample; //The audio buffer.
  pBuffer = ^tBuffer; //Pointer to the audio buffer.
  NoteRec = record
        Peak   : Integer;
        Note1  : Char;
        Note2  : Char;
        Octave : Integer;
        E1     : Integer;
        A      : Integer;
        D      : Integer;
        G      : Integer;
        B      : Integer;
        E2     : Integer;
      end;{of record}

type
  TMain = class(TForm) //Player
    MainMenu1: TMainMenu; //Menu
    File1: TMenuItem; //Top level menu item
    Open1: TMenuItem; //Second level menu item
    Exit1: TMenuItem; //Exit in menu
    OpenDialog1: TOpenDialog; //Multimedia status
    Timer1: TTimer; //Multimedia progress gauge
    OpenCD1: TMenuItem; //Open CD menu item
    FFT1: TFastFourier;
    View1: TMenuItem;
    ShowInput1: TMenuItem;
    Help1: TMenuItem;
    About1: TMenuItem;
    HowtoUseHelp1: TMenuItem;
    Guide1: TMenuItem;
    SearchforHelpOn1: TMenuItem;
    Procedures1: TMenuItem;
    Index1: TMenuItem;
    Contents1: TMenuItem;
    SaveDialog1: TSaveDialog;
    PageControl1: TPageControl;
    NoteSampler: TTabSheet;
    Gauge1: TGauge;
    Memo1: TMemo;
    NoteBox: TEdit;
    Sample: TButton;
    WriteB: TButton;
    Button1: TButton;
    Button2: TButton;
    FilePlayer: TTabSheet;
    Gauge2: TGauge;
    MediaPlayer1: TMediaPlayer;
    Edit1: TEdit;
    ChordRef: TTabSheet;
    ComboBox1: TComboBox; //Just a label
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ExitClick(Sender: TObject);
    procedure SampleClick(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure OpenCD1Click(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure ShowInput1Click(Sender: TObject);
    procedure WriteBClick(Sender: TObject);
    procedure Fretboard1Click(Sender: TObject);
    procedure ComboBox1Change(Sender: TObject);
    procedure Contents1Click(Sender: TObject);
    procedure SearchforHelpOn1Click(Sender: TObject);
    procedure HowtoUseHelp1Click(Sender: TObject);
    procedure Index1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure Guide1Click(Sender: TObject);
    procedure ClearTab(Sender: TObject);
    procedure Button2Click(Sender: TObject);
  private
    { Private declarations }
    p_Wave_Hdr  : pWaveHdr;//Wave Header
    p_Buffer    : pBuffer;//Input Buffer
    hWave_Hdr   : hGlobal;
    hBuffer     : hGlobal;
    Buffer_Bytes : integer;
    hWave_In    : hWaveIn;
    pcm         : tWaveFormatEx;
    WaveCaps    : pWaveInCaps;
    Recording   : boolean;
    Shutoff     : boolean;
    Closing     : boolean;
    AudioShow   : boolean;
    FretShow    : boolean;
    Fund : Integer;
    CurrentNote : NoteRec;
    Level : Integer;
    Writing : Boolean;

    //Private procedures
    procedure StopRecording;
    procedure StartRecording;

    //Call Backs
    procedure MMInOpen(var msg: Tmessage);message MM_WIM_OPEN;
    procedure MMInClose(var msg: Tmessage);message MM_WIM_CLOSE;
    procedure MMInData(var msg: Tmessage);message MM_WIM_DATA;


  public
    { Public declarations }
  end;

var
  Main: TMain;


implementation

uses AudUnit, FretUnit, Unit1, AbUnit, WaveInfo;

{$R *.DFM}

procedure GetNote;
Var
  DataFile: TextFile;
  Tab : Char;
  found : boolean;
  PreviousNote, Temp, LastNote : NoteRec;
  PrevString, CurrString : string;
begin
  //Open the notes.dat lookup file uses IO checking to see if the open worked
  AssignFile(DataFile, 'Notes.dat');
  {$I-}
  Reset(DataFile);
  {$I+}
  if IOResult = 0 then
  begin
    found:=false;
    Temp.Note1:='C';
    Temp.Note2:='#';
    //Loop checks to see if the note being read has a larger peak than the
    //previous note read.  If it does then the note being sampled is taken
    //to be the previous note - gives a note a range of possible values.
    while not found and not EOF(DataFile) do
    begin
      LastNote:=Temp;
      ReadLn(DataFile,Temp.Peak,Tab,Temp.Note1,Temp.Note2,Tab,
           Temp.Octave,Tab,Temp.E2,Tab,Temp.B,Tab,Temp.G,Tab,
           Temp.D,Tab,Temp.A,Tab,Temp.E1);
      If Main.Fund = Temp.Peak then
        found:=true;
      If Main.Fund > Temp.Peak then
      begin
        Temp:=LastNote;
        found:=true;
      end;
    end;
    //Sets the current note variable
    Main.CurrentNote:=Temp;
    CurrString := Temp.Note1+Temp.Note2+inttostr(temp.octave);
    //Updates the memo control with the current note if required
    if main.writing then
      with main.CurrentNote do
        Main.Memo1.Lines.Add(Note1+Note2+' '+inttostr(Octave)+' '+Inttostr(E2)
                    +' '+inttostr(B)+' '+inttostr(G)+' '+inttostr(D)+' '
                    +inttostr(A)+' '+inttostr(E1));
  end
  else
    MessageDlg('The file notes.dat not found.  Note sampling not available'
                , mtWarning, [mbOk], 0);
  CloseFile(DataFile);

end;

//Updates the volume unit gauge, with colour changes dependent upon level
procedure UpdateVuMeter;
begin
  case Main.Level of
    0..30:
      main.gauge1.forecolor:=clLime;
    31..60:
      main.gauge1.forecolor:=clYellow;
    61..128:
      main.gauge1.forecolor:=clRed;
  end; {of case}
  main.gauge1.progress:=Main.Level;
end;

//Sets the edit control to show the current note
procedure UpdateNote;
begin
  with Main.CurrentNote do
  begin
    if Note2 = '#' then
      Main.NoteBox.Text:=Note1 + Note2 + inttostr(Octave)
    else
      Main.NoteBox.Text:=Note1 + inttostr(Octave);
  end
end;

//Updates the finger board according to current values.  An if for each string
procedure UpdateFretboard;
begin
  Fretboard.refresh;
  Fretboard.Canvas.Pen.Color:=clRed;
  with Main.CurrentNote do
  begin
    If E2 <> 99 then
      fretboard.canvas.Ellipse(2+(round(((fretboard.width-30)/21))*E2),
              16,
              10+(round(((fretboard.width-30)/21))*E2),
              24);
    If B <> 99 then
      fretboard.canvas.Ellipse(2+(round(((fretboard.width-30)/21))*B),
              11+(round((Fretboard.Height-60)/5)),
              10+(round(((fretboard.width-30)/21))*B),
              19+(round((FretBoard.Height-60)/5)));
    If G <> 99 then
      fretboard.canvas.Ellipse(2+(round(((fretboard.width-30)/21))*G),
              11+(round((Fretboard.Height-60)/5)*2),
              10+(round(((fretboard.width-30)/21))*G),
              19+(round((FretBoard.Height-60)/5))*2);
    If D <> 99 then
      fretboard.canvas.Ellipse(2+(round(((fretboard.width-30)/21))*D),
              11+(round((Fretboard.Height-60)/5)*3),
              10+(round(((fretboard.width-30)/21))*D),
              19+(round((FretBoard.Height-60)/5))*3);
    If A <> 99 then
      fretboard.canvas.Ellipse(2+(round(((fretboard.width-30)/21))*A),
              11+(round((Fretboard.Height-60)/5)*4),
              10+(round(((fretboard.width-30)/21))*A),
              19+(round((FretBoard.Height-60)/5))*4);
    If E1 <> 99 then
      fretboard.canvas.Ellipse(2+(round(((fretboard.width-30)/21))*E1),
              Fretboard.Height-54,
              10+(round(((fretboard.width-30)/21))*E1),
              Fretboard.Height-46);
  end;
end;

//Callback for the wave open message.
procedure TMain.MMInOpen(var msg: Tmessage);
var
  status : MMRESULT;
begin
  //populate the wave header
  with p_wave_hdr^ do
  begin
    lpData := pchar(p_Buffer);
    dwBufferLength := BufferLength;
    dwBytesRecorded := BufferLength;
    dwUser := 0;
    dwFlags := 0;
    dwLoops := 1;
    reserved :=0;
  end;

  //prepare wave header
  status:=WaveInPrepareHeader(hWave_in,p_wave_hdr,SizeOf(tWaveHdr));
  if status <> 0 then
  begin
    MessageDlg('There was an error preparing a wave header.  Check hardware.'
                , mtError, [mbOk], 0);
    exit;
  end;

  //send the header to the driver - sent to prevent data loss
  status:=WaveInAddBuffer(hWave_In,p_wave_hdr,SizeOf(tWaveHdr));
  if status <> 0 then
  begin
    MessageDlg('There was an error adding a wave buffer.  Check hardware.'
                , mtError, [mbOk], 0);
    exit;
  end;

end;

//Callback function for the wave device close message
procedure TMain.MMInClose(var msg: Tmessage);
var
  status : MMRESULT;
begin
  //unprepare wave header1
  status:=WaveInUnPrepareHeader(hWave_in,p_wave_hdr,SizeOf(tWaveHdr));
  if (status <> 0) and Recording then
  begin
    MessageDlg('There was an error unpreparing a wave header.  Check hardware.'
                , mtError, [mbOk], 0);
    exit;
  end;

end;

//Deals with the audio data as it arrives.
procedure TMain.MMinData(var msg: Tmessage);
var
  status : MMRESULT;
  count, sum : integer;
  HzPeak1, HzPeak2, HzPeak3 : integer;
  Peak1Val, Peak2Val, Peak3Val : double;
  data, y : real;
begin
  //Clear the fft component ready fo new data
  FFT1.Clear;
  //Clear and initialise the audion input graph
  AudioInput.InputGraph.ClearGraf;
  AudioInput.InputGraph.MoveTo(0,128);
  sum:=0;
  HzPeak1:=0; HzPeak2:=0; HzPeak3:=0; Peak1Val:=0; Peak2Val:=0; Peak3Val:=0;
  //Loop through the audio buffer
  for count:=0 to BufferLength-1 do
  begin
    //set data variable to the audio data value
    data:=longint(p_wave_hdr.lpdata[count]);
    sum:=sum+round(abs(data-128));
    //Clear the imaginary spectrum of the fft component
    FFT1.ImagSpec[count]:=0;
    //Fill in the real spectrum of the fft component
    FFT1.RealSpec[count]:= data;
    //Update the audio input graph
    If AudioShow then
      AudioInput.InputGraph.DrawTo(count,data);
  end;
  If Audioshow then
    AudioInput.InputGraph.ShowGraf;
  //Update the level meter
  Level:= round(sum/BufferLength);
  UpdateVUMeter;

  //do the FFT
  FFT1.Transform;
  //Find the frequency peaks using some simple sorts
  for count:=Threshold to Cutoff do
  begin
    y:=FFT1.PowerSpec[count];
    If (y > Peak1Val) then
    begin
      Peak3Val:=Peak2Val;
      HzPeak3:=HzPeak2;
      Peak2Val:=Peak1Val;
      HzPeak2:=HzPeak1;
      Peak1Val:=y;
      HzPeak1:=count-1;
    end
    else if (y > Peak2Val) then
    begin
      Peak3Val:=Peak2Val;
      HzPeak3:=HzPeak2;
      Peak2Val:=y;
      HzPeak2:=count-1;
    end
    else if (y > Peak3Val) then
    begin
      Peak3Val:=y;
      HzPeak3:=count-1;
    end

  end;

  //Find the fundamental - the lowest significant peak
  Fund:=999;
  if HzPeak1 < Fund then
    Fund:=HzPeak1;
  if HzPeak2 < Fund then
    Fund:=HzPeak2;
  if HzPeak3 < Fund then
    Fund:=HzPeak3;

  //Sweet is a constant.  By setting this threshold and not updating the
  //display if the peak value is lower, notes in decay are ignored giving
  //a more useful result.
  if Peak1Val > sweet then
  begin
    GetNote;
    UpdateNote;
    If FretShow then
      UpdateFretboard;
  end
  else
    NoteBox.Text:=' ';
  //Add the next wave buffer to be filled.
  status:=WaveInAddBuffer(hWave_In,p_wave_hdr,SizeOf(tWaveHdr));
  if (status <> 0) and Recording then
  begin
    MessageDlg('There was an error adding a wave buffer.  Check hardware.'
                , mtError, [mbOk], 0);
    StopRecording();
    exit;
  end;
end;

procedure TMain.FormCreate(Sender: TObject);
var
  loop   : integer;
  status : mmresult;
begin
  //Check wave device capabilities
  status := waveInGetDevCaps(0,WaveCaps,0);
  if status <> 0 then
  begin
    MessageDlg('There are no wave devices on this PC.  Check hardware.'
                , mtError, [mbOk], 0);
    exit;
  end;
  //Get the chord list from file into the combo.
  if FileExists('Chordlst.dat') then
    ComboBox1.Items.LoadFromFile('Chordlst.dat')
  else
    MessageDlg('Chordlst.dat not present.  Chord database will not work.'
                , mtError, [mbOk], 0);
  FFT1.SpectrumSize:=SpectrumSize;
  //Get the memory for the wave header
  hWave_Hdr := GlobalAlloc (gHnd or gMem_Share, SizeOf(tWaveHdr));
  p_Wave_Hdr := pWaveHdr (GlobalLock(hWave_Hdr));

  //Set input buffer size
  Buffer_Bytes := BufferLength;

  //Get memory for input header
  hBuffer := GlobalAlloc(gHnd or gMem_Share, Buffer_Bytes);
  p_Buffer := pBuffer (GlobalLock(hBuffer));

  hWave_In := 0;

  //Set state variables
  ShutOff := false;
  Closing := false;
  Recording := false;
end;

procedure TMain.FormDestroy(Sender: TObject);
begin
  shutoff := true;
  if writing then
    writing:=false;
  //Free memory resources
  globalunlock (hWave_hdr);
  globalfree (hwave_hdr);
  globalunlock (hbuffer);
  globalfree (hbuffer);
end;

procedure TMain.ExitClick(Sender: TObject);
begin
  if recording then stoprecording;
  close;
end;

procedure TMain.StopRecording;
begin
  if Recording then
  begin
    shutoff:=true;
    //Close down the wave device.
    WaveInStop (hWave_In);
    WaveInReset(hWave_In);
    WaveInClose(hWave_In);
    Recording := false;
  end;
end;

//Start the sampling process
procedure TMain.StartRecording;
var
  status : MMRESULT;
  code : integer;
begin
  if recording then stoprecording;
  recording:=true;
  sample.caption := 'STOP';  //change the caption of the start button

  //fill in the WaveFormatex structure to set recording paramaters
  with pcm do
  begin
    wFormatTag      := WAVE_FORMAT_PCM;
    nChannels       := 1;
    nSamplesPerSec  := SampleRate;
    nAvgBytesPerSec := SampleRate*(round(BitsPerSample/8));
    nBlockAlign     := round(BitsPerSample/8);
    wBitsPerSample  := BitsPerSample;
    cbSize          := 0;
  end;
  shutoff:=false;

  //Try and open the wave input device for recording
  status := WaveInOpen(@hWave_In, 0, @pcm, Handle, 0, callback_window);
  if status <> 0 then
  begin
    MessageDlg('There was an error opening the wave device.  Check hardware.'
                , mtError, [mbOk], 0);
    exit;
  end;

  //start  recording
  status := WaveInStart(hWave_In);
  if status <> 0 then
  begin
    MessageDlg('There was an error starting recording.  Check hardware.'
                , mtError, [mbOk], 0);
    exit;
  end;
end;

procedure TMain.SampleClick(Sender: TObject);
begin
  If Recording then
  begin
    sample.caption := '&Sample';
    StopRecording
  end
  else StartRecording;
end;

//Controls the media player
procedure TMain.Open1Click(Sender: TObject);
begin
  with opendialog1 do
  begin
    gauge2.Visible:=true;
    Edit1.Text:='';
    if execute then
    begin
      mediaplayer1.filename:=filename;
      mediaplayer1.autoopen:=false;
      mediaplayer1.devicetype:=dtAutoSelect;
      mediaplayer1.shareable:=false;
      Edit1.Text:='Loaded ' + filename;
    end;
    with mediaplayer1 do
    begin
      open;
      if mediaplayer1.error =0 then
      begin
        gauge2.maxvalue := length;
        gauge2.progress := 0;
        timer1.enabled := true;
        frames:=1;
        step;
      end
      else
        Edit1.Text:='Cannot Open';
    end;
  end;
end;

//Update the media player display - timer event
procedure TMain.Timer1Timer(Sender: TObject);
var
  Trk, Min, Sec : Word;
  dwReturn : dword;
  iNumTracks : integer;
begin
  if mediaplayer1.devicetype=dtCDAudio then
  begin
    with mediaplayer1 do
    begin
      Trk := MCI_TMSF_TRACK(Position);
      Min := MCI_TMSF_MINUTE(Position);
      Sec := MCI_TMSF_SECOND(Position);
      if trk > 25 then
        Edit1.Text:=('No CD in drive')
      else
        Edit1.Text:=('Track '+inttostr(Trk)+ ' of ' + inttostr(tracks) +
                     ', Track Time ' + inttostr(Min) + ':' + inttostr(Sec));
    end;
  end
  else
    gauge2.progress := mediaplayer1.position;
end;

//Opens an audio CD for playing
procedure TMain.OpenCD1Click(Sender: TObject);
begin
  with mediaplayer1 do
  begin
    gauge2.Visible := false;
    filename:='';
    autoopen:=true;
    devicetype:=dtCDAudio;
    shareable:=true;
    open;
    if mediaplayer1.error = 0 then
    begin
      gauge2.maxvalue := length;
      gauge2.progress := 0;
      timer1.enabled:=true;
    end
    else
      Edit1.Text:='Cannot Open - No CD?';
  end;
end;

procedure TMain.Exit1Click(Sender: TObject);
begin
  Close;
end;

//Shows the audio input graph
procedure TMain.ShowInput1Click(Sender: TObject);
begin
  AudioShow:=true;
  AudioInput.Show;
end;

//Starts writing notes to memo control
procedure TMain.WriteBClick(Sender: TObject);
begin
  if WriteB.Caption = '&Write to Text' then
  begin
    WriteB.Caption := 'Stop &Writing';
    Writing := true;
  end
  else
  begin
    WriteB.Caption := '&Write to Text';
    writing := false;
  end;
end;

//Shows the fretboard
procedure TMain.Fretboard1Click(Sender: TObject);
begin
  FretShow:=True;
  Fretboard.Show;
end;

//Shows the chord chosen in the combo box on the fretboard
procedure TMain.ComboBox1Change(Sender: TObject);
var
  ChordFile : TextFile;
  Found : boolean;
  ItemNo,n : Integer;
  Tab : char;
  Chord : string;
begin
  If Recording then
  begin
    sample.caption := '&Sample';
    StopRecording;
  end;
  AssignFile(ChordFile, 'Chord.dat');
  {$I-}
  Reset(ChordFile);
  {$I+}
  if IOResult = 0 then
  begin
    Found:=false;
    For n:=0 to ComboBox1.ItemIndex -1 do
      Readln(ChordFile);
    with main.currentnote do
      Readln(ChordFile,E2,tab,B,tab,G,tab,D,tab,A,tab,E1,tab,Chord);
    CloseFile(ChordFile);
    Fretboard.Show;
    UpdateFretboard;
  end
  else
  begin
      MessageDlg('File chord.dat not found.', mtError, [mbOk], 0);
      Exit;
  end;
end;


//Help file links
procedure TMain.Contents1Click(Sender: TObject);
begin
  Application.HelpCommand(HELP_CONTENTS, 0);
end;

procedure TMain.SearchforHelpOn1Click(Sender: TObject);
begin
  Application.HelpCommand(HELP_PARTIALKEY,0);
end;

procedure TMain.HowtoUseHelp1Click(Sender: TObject);
begin
  Application.HelpCommand(HELP_HELPONHELP, 0);
end;

procedure TMain.Index1Click(Sender: TObject);
begin
  Application.HelpCommand(HELP_FINDER, 0);
end;

procedure TMain.Guide1Click(Sender: TObject);
begin
  Application.HelpCommand(HELP_CONTEXT, 300);
end;


//Shows the about box
procedure TMain.About1Click(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

//Clears the memo box containing the sampled notes
procedure TMain.ClearTab(Sender: TObject);
var
  l,s : integer;
begin
  Main.Memo1.Lines.Clear;
  Main.Memo1.Lines.Add('Sampled Notes');
end;

//Saves the contents of the memo box to a file
procedure TMain.Button2Click(Sender: TObject);
begin
  with SaveDialog1 do
    if Execute then
      Memo1.Lines.SaveToFile(Filename);
end;

end.
